
/*
    middleware.c	from Linux 4D Project (http://linux-4d.sourceforge.net)
    
    Author:		Matteo Fasanella (mfasanella@deis.unibo.it)
    
*/



#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>

#include "festival.h"
#include "parser.h"
#include <sys/myspeakup.h>
#include "xcatch.h"
#include "xinterceptor.c"

#define firstCon	1
#define lastCon		49
#define XCon		10


int init_all();
int close_all();
static void sig_usr1(int);
static void sig_usr2(int);
void gestCons(int);
void gestFn(char *, int);
char *makeAcc(char *);

int fd1, fd2;
int cons=0;
char *fmState="";

char *char_ptr, *keys, *saved;


struct domain *Fnbase[50];	// max 50 Console + applicaz. X
struct head *Fncurr[50];



int main (void) {

    init_all();
    XInitAll();
    
    while (1) {
        XNextEvent(disp, &xev);
        procEvent();
    }
    
    close_all();

return 0;
}





int procEvent() {
    Action action;
    XTextProperty *prop;
    char tmp[128]="",  *keys, *self;
    int i, x=0, y=0;


    if (set[xev.type].is_set) {
	action = set[xev.type].default_action;
        prop = (XTextProperty *) malloc(sizeof(XTextProperty)); bzero(prop, sizeof(XTextProperty));	
	if (NULL!=action) {
	    XGetTextProperty(disp, xev.xany.window, prop, XA_WM_NAME);
	    if (prop->value != NULL) {
		if (strcmp(prop->value, "kwin") != 0) {
    		    xcurrapp->win = xev.xany.window; xcurrapp->app = prop->value;
		    printf("Event: %d,\twindow: %lu,\t\tname: %s\n", xev.type, xev.xany.window, prop->value);		
//		    printf("\t\txcurrapp->win: %lu,\txcurrapp->app: %s\n", xcurrapp->win, xcurrapp->app);		
		}
	    }
	    action(action, xev);
	}
	free(prop);
    }

    if (xev.type==CreateNotify) {
	XCreateWindowEvent *xcwe;
	xcwe=(XCreateWindowEvent *)&xev;
	select_all_windows(disp, xcwe->window, type);
    }

    if (xev.type==KeyPress) {
	    // non riesce a gestire il CapsLock (servirebbe un contatore o un booleano)
        keys = (char *) malloc(32);
	XQueryKeymap(disp, keys);
	for (i=0; i<32*8; i++) {
	    register char *str;
    	    str = (char *) KeyCodeToStr(i, BIT(keys, i), KeyModifiers(keys));
    	    if (BIT(keys, i)!=0) {
		strcat(tmp, str);
	    }
	}
	self = (char *) malloc(128);   
	if ( (strlen(tmp) == 1) || (tmp[0]='^')) sprintf(self, "Self%s", tmp);
	else strcpy(self, tmp);
        printf("Event: %d, keycode: %d, Key: %s, Self: %s\n", xev.type, xev.xkey.keycode, tmp, self);
	
	gestFn(self, 10);
    }




return 0;
}



int init_all() {

    char *server=0, filetmp[32]="", tmp[128]="";
    int port=-1, i=0;


    // init signal handlers
    if (signal(SIGUSR1, sig_usr1) == SIG_ERR) { printf("can't catch SIGUSR1\n"); exit(-1); }
    if (signal(SIGUSR2, sig_usr2) == SIG_ERR) { printf("can't catch SIGUSR2\n"); exit(-1); }


    //  init log files
    fd1=open("text", O_RDWR|O_CREAT, 00666);
    fd2=open("ctrl", O_RDWR|O_CREAT, 00666);


    // init festival 
    info = festival_default_info();
    if (server != 0)  info->server_host = server;
    if (port != -1)  info->server_port = port;
//    printf("server: %s, on port: %d, with mode: %s\n", info->server_host, info->server_port, info->text_mode);
    info = festivalOpen(info);

    
    // init FileManager
    for (i=firstCon; i<=lastCon; i++) {
	bzero(filetmp, 32);
	sprintf(filetmp, "rules/f%d.rul", i);
        Fnbase[i] = create_rules(filetmp);
	if (Fnbase[i] != NULL) {
	    Fncurr[i] = Fnbase[i]->headptr;
	    printf("init console %d, with: %s => %s\n", i, filetmp, Fnbase[i]->con_name); 
//   		print_rules(Fnbase[i]);
	} 
    }


return 0;
}


int close_all() {

    festivalClose(info);
    close(fd1);
    close(fd2);    

return 0;
}


static void sig_usr1(int signo) {

    char *ritorno=(char *) malloc(8192);
    int i=0;

    if (signo==SIGUSR1) {
	get2speak(ritorno);   
        printf("get2speak: %s\n", ritorno);

	write (fd1, ritorno, strlen(ritorno));	
	if (festivalSpeech(info, ritorno) != 0) printf("Errore nella Speech!\n");    	
    }
    
    return;
}



static void sig_usr2(int signo) {

    char *ritorno=(char *) malloc(8192);
    int i=0;

    if (signo==SIGUSR2) {
	get2ctrl(ritorno);   
	printf("get2ctrl: %s\n", ritorno);
	
	printf("... %d\n", ritorno[4]);
	if ( (ritorno[4] >= 1) && (ritorno[4] <= 27) ) {	// per i '^...'
	    if ( ritorno[4] == 27)  sprintf(ritorno, "Self^[");	    
	    else  sprintf(ritorno, "Self^%c", 'A'+ritorno[4]-1);	    
	}
        if ( ritorno[4] == 127)  sprintf(ritorno, "Self^<");	    
	
	
//	printf("ritorno: %s\n", ritorno);
	write (fd2, ritorno, strlen(ritorno));	

	// se c'e` cambio di console (=> voce di sistema)
	if (strncmp(ritorno, "Cons", 4) == 0) {
	    ((strlen(ritorno)==5)?(cons = (ritorno[4] - '0')):(cons = ((ritorno[4] - '0')*10)+(ritorno[5] - '0')) );
	    gestCons(cons);
	    return;
	}

	// se c'e` evento ad un servizio 
	if ( (cons >= firstCon) && (cons <= lastCon) ) {
	    gestFn(ritorno, cons);	// e lo passa a gestFn
	    return;
	}

	// altrimenti, se e` su una console sconosciuta legge il comando
	if (festivalSpeech(info, ritorno) != 0) printf("Errore nella Speech!\n");    	
    }
    
    return;
}



void gestCons(int consn) {

    char toRead[1024]="";

    sprintf(toRead, "terminale..... %d..", consn);

    if ( (consn >= firstCon) && (consn <= lastCon) ) {
	if (Fnbase[consn] == NULL) { 
	    printf("%s\n", toRead);
    	    if (festivalCommand(info, "(voice_lp_mbrola)\n") != 0) printf("Errore nella Speech!\n");    		    	
    	    if (festivalSpeech(info, toRead) != 0) printf("Errore nella Speech!\n");    		    
	    return;
	}
	if (strcmp(Fnbase[consn]->con_name, "") == 0) strcpy(Fnbase[consn]->con_name, toRead);
	printf("%s\n", Fnbase[consn]->con_name);
        if (festivalCommand(info, "(voice_lp_mbrola)\n") != 0) printf("Errore nella Speech!\n");    		    	
//	if (festivalCommand(info, "(Parameter.set \"Duration_Stretch\" 1.2)\n") != 0) printf("Errore nella Speech!\n");    		    	

        if (festivalSpeech(info, makeAcc(Fnbase[consn]->con_name)) != 0) printf("Errore nella Speech!\n");    		    
	return;
    }

    printf("Console %d\n", consn);
    if (festivalSpeech(info, toRead) != 0) printf("Errore nella Speech!\n");    	

return;
}




void gestFn(char *event, int con) {

struct cell *currcell=NULL;
struct action *curract=NULL;
int esci=0, x=firstCon, y=lastCon;
char toRead[1024]="";
struct domain currdom;



//    printf("con: %d, XCon: %d\n", con, XCon);
    if (con == XCon) {	// server X: ogni prog. ha una consolle da Fn20 in su`
			// cerca il num di consolle associato al prog (dal
			// file di regole), ed inposta "con"
	for (x=firstCon; x<=lastCon; x++) {
//	    x=21;
	    if ((Fnbase[x] != NULL) && (strcmp(Fnbase[x]->con_name, xcurrapp->app) == 0))  {
		con = x;
		printf("%s %s, strcmp=%d\n\n", Fnbase[x]->con_name, xcurrapp->app, strcmp(Fnbase[x]->con_name, xcurrapp->app));
	    }
	}
	if (con == XCon) { printf("Non trovo il file di regole per %s\n", xcurrapp->app); return;}
    }


    if (Fnbase[con] == NULL) { 
        if (festivalSpeech(info, "Nessun file di regole per questo terminale!") != 0) printf("Errore nella Speech!\n");    		    
	printf("Nessun file di regole per questo terminale!\n"); 
	return;
    }

    if (festivalCommand(info, "(voice_pc_mbrola)\n") != 0) printf("Errore nella Speech!\n");    		    	
    if (festivalCommand(info, "(Parameter.set \"Duration_Stretch\" 1.1)\n") != 0) printf("Errore nella Speech!\n");    		    	

    if (strcmp(event, "Self ") == 0) strcpy(event, "Self"); // per lo spazio

    currcell=Fncurr[con]->table;

			// il controllo su "Char" e` da ottimizzare
//    if ( (strncmp(event, "Self", 4) == 0) && ( (event[5]>='a' && event[5]<='z')||(event[5]>='A' && event[5]<='Z')) ) {	
    if ( (strncmp(event, "Self", 4) == 0) && ( ((event[4]>='a') && (event[4]<='z'))) ) {	
//	printf("forse e` Char\n");
        while ( (currcell != NULL) && (!esci) ) {
//    	    printf("currcell->event: %s, event:%s, strcmp:%d\n", currcell->event, event, strcmp(currcell->event, event));
    	    if (strcmp(currcell->event, "Char") == 0) esci=1;
	    else if (currcell->next_cell != NULL)
		currcell=currcell->next_cell;
	    else esci=1;
	}
	
	esci = 0;
	if (strcmp(currcell->event, "Char") == 0)  goto prova;
	    
    } 

    currcell=Fncurr[con]->table;
    
    while ( (currcell != NULL) && (!esci) ) {
//		printf("currcell->event: %s, event:%s, strcmp:%d\n", currcell->event, event, strcmp(currcell->event, event));
    	    if (strcmp(currcell->event, event) == 0) esci=1;
	    else if (currcell->next_cell != NULL)
		currcell=currcell->next_cell;
	    else esci=1;
	}
	if (strcmp(currcell->event, event) != 0) {
    	    if (festivalSpeech(info, "Evento sconosciuto....") != 0) printf("Errore nella Speech!\n");    		    
	    printf("Unknown event, state unchanged\n");
	    return;
	}

    prova:

    curract = currcell->actions;
    while ( curract != NULL )  {

//	printf("event, action: %s, %s\n", currcell->event, curract->action);    
	if (strcmp(curract->action, "nil") == 0) return;
    
        if (strcmp(curract->action, "go") == 0) {	// evento "go"
    	    Fncurr[con]=curract->next_table;
	    sprintf(toRead, "go.. to.. %s", Fncurr[con]->header);
//    	    if (festivalSpeech(info, makeAcc(toRead)) != 0) printf("Errore nella Speech!\n");    		    
	    printf("go to: [%s], toRead:%s\n", Fncurr[con]->header, makeAcc(toRead));

    
    	    currcell=Fncurr[con]->table;
//	    while (currcell != NULL) {
//	    	printf("currcell->event: %s\n", currcell->event);
//		currcell=currcell->next_cell;
//	    }
	
	} else if (strcmp(curract->action, "SayText") == 0) {
	    sprintf(toRead, "%s....", curract->parms);
    	    if (festivalSpeech(info, makeAcc(toRead)) != 0) printf("Errore nella Speech!\n");
	    printf("SayText: %s \n", makeAcc(toRead));		
		
	} else {  	// azioni da compiere (eventualmente su "parms")
	    sprintf(toRead, "%s..... %s...", curract->action, curract->parms);
	    printf("%s(%s), toRead:%s\n", curract->action, curract->parms, makeAcc(toRead));
    	    if (festivalSpeech(info, makeAcc(toRead)) != 0) printf("Errore nella Speech!\n");    		    	
	}
    
	curract = curract->next_action;
	
    }	// end while    

return;
}



char *makeAcc(char *toacc) {  // make a phrase, readable

int i=0;

    while (i<strlen(toacc)) {
	if (!(((toacc[i]>='a') && (toacc[i]<='z')) || ((toacc[i]>='A') && (toacc[i]<='Z')) || (toacc[i]=='.') || ((toacc[i]>='0') && (toacc[i]<='9')) ) )
	    toacc[i] = ' ';
	i++;
    }
    
return toacc;
}
	

