
#include <X11/Xlib.h>
#include <X11/X.h>

#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include <string.h>
#include <stdio.h>

Display *disp;
XEvent xev;
unsigned long type;

typedef struct Xapp {
    unsigned long int win;
    char *app;
} Xapp;

Xapp *xcurrapp;


typedef int (*Action)();

typedef struct SetTable {
	int is_set;
	Action default_action;
} SetTable;

SetTable set[(LASTEvent)];	// LASTEvent <=> MAX_EVENT


void XInitAll();
Action	ActionCall(XEvent *xev, int mode);
void	select_all_windows(Display *disp, Window root, unsigned long type);
int PrintEvent (Action *action, XEvent xev);
int NewXErrorHandler(Display *disp, XErrorEvent *xeev);
int NewXIOErrorHandler(Display *disp);





void XInitAll() {
    int *tmp;
    XTextProperty *prop;
    Window *currw;
    

    disp=XOpenDisplay(":0");
    if (disp==NULL) { fprintf(stderr, "Cannot open X display :0\n"); exit(1); }
    XSetErrorHandler  (NewXErrorHandler);
    XSetIOErrorHandler(NewXIOErrorHandler);

    bzero(set, sizeof(set));
    set[FocusIn].is_set	= 1; 	set[FocusIn].default_action = &PrintEvent;
    set[FocusOut].is_set = 1; 	set[FocusOut].default_action = &PrintEvent;
    set[KeyPress].is_set = 1; 	set[KeyPress].default_action = &PrintEvent;
 
    type=FocusChangeMask|SubstructureNotifyMask|KeyPressMask;

    XGrabServer(disp);
    select_all_windows(disp, DefaultRootWindow(disp), type);
    XUngrabServer(disp);
    
    xcurrapp = (Xapp *) malloc(sizeof(struct Xapp));
    
    currw = (Window *) malloc(sizeof(Window));
    XGetInputFocus(disp, currw, tmp);
    prop = (XTextProperty *) malloc(sizeof(XTextProperty)); bzero(prop, sizeof(XTextProperty));	
    XGetTextProperty(disp, *currw, prop, XA_WM_NAME);
    xcurrapp->win = *currw; 	xcurrapp->app = prop->value;
    printf("Initial Values\txcurrapp->win: %lu,\txcurrapp->app: %s\n", xcurrapp->win, xcurrapp->app);		
        
    free(prop);
    free(currw);

    
}




void select_all_windows(Display *disp, Window root, unsigned long type) {
	Window	parent, *children;
	int	stat, i, j, k;
	unsigned int	nchildren;

	XSelectInput(disp, root, type);

	stat=XQueryTree(disp, root, &root, &parent, &children, &nchildren);
	if (stat == 0) { fprintf(stderr, "Can't query window tree.\n"); return; }
	if (nchildren == 0) return;

	for(i=0; i < nchildren; i++) {
		select_all_windows(disp, children[i], type);
		}     

	XFree((char *)children);
	}




int PrintEvent(Action *action, XEvent xev) {
//    printf("Event: %d\n", xev.type);
return;
}



int NewXErrorHandler(Display *disp, XErrorEvent *xeev) {  // Normal Errors 
	char buf[65];
	
	XGetErrorText(disp, xeev->error_code, buf, 65);
	fprintf(stderr, "Xerror: %s(%d) OpCode: %d.%d Display: \"%s\" RID: %d\n", buf, xeev->error_code, xeev->request_code, xeev->minor_code, DisplayString(disp), xeev->resourceid);
	fflush(stdout);
}


	
                             
int NewXIOErrorHandler(Display *disp) {  // Fatal Errors 
	fprintf(stderr,"Xerror: Fatal error on display %s.\n",DisplayString(disp));
	fflush(stdout);
}




