#define BIT(c, x)   ( c[x/8]&(1<<(x%8)) )

#define KEYSYM_STRLEN   64

#define SHIFT_DOWN  1
#define LOCK_DOWN   2
#define CONTROL_DOWN    3
#define SHIFT_INDEX 1  /*index for XKeycodeToKeySym(), for shifted keys*/


char *KeyCodeToStr(int code, int down, int mod);



struct conv {char from[20], to[5];} conv_table[] = {
   /* shift & control replaced with nothing, since they are appearent
      from the output */
   {"return","\n"},    {"escape","^["},    {"delete", "^H"},
   {"shift",""},       {"control",""},     {"tab","Tab"},
   {"space", " "},     {"exclam", "!"},    {"quotedbl", "\""}, 
   {"numbersign", "#"},{"dollar", "$"},    {"percent", "%"},
   {"ampersand", "&"}, {"apostrophe", "'"},{"parenleft", "("}, 
   {"parenright", ")"},{"asterisk", "*"},  {"plus", "+"},
   {"comma", ","},     {"minus", "-"},     {"period", "."},    
   {"slash", "/"},     {"colon", ":"},     {"semicolon", ";"}, 
   {"less", "<"},      {"equal", "="},     {"greater", ">"},   
   {"question", "?"},  {"at", "@"},        {"bracketleft", "["},
   {"backslash", "\\"},{"bracketright", "]"},{"asciicircum", "^"},   
   {"underscore", "_"},{"grave", "`"},     {"braceleft", "{"}, 
   {"bar", "|"},       {"braceright", "}"},{"asciitilde", "~"},    
   {"",""}
};


int StrToChar(char *data) {
   int i=0;
   while (conv_table[i].from[0]!=0 || conv_table[i].to[0]!=0) {
      if (!strncasecmp(data,conv_table[i].from, strlen(conv_table[i].from)) ) {
         strcpy(data,  conv_table[i].to);
         return 1;
      }
      i++;
   }
   return 0;
}


char *KeyCodeToStr(int code, int down, int mod) {
   static char *str, buf[KEYSYM_STRLEN+1];
   int index;
   KeySym  keysym;

   /* get the keysym for the appropriate case */
   if (mod==SHIFT_DOWN) index=SHIFT_INDEX;
   else index=0;
   keysym=XKeycodeToKeysym(disp, code, index);
   if (NoSymbol==keysym) return "";

   /* convert keysym to a string, copy it to a local area */
   str=XKeysymToString(keysym);
   if (NULL==str) return "";
   strncpy(buf, str, KEYSYM_STRLEN); buf[KEYSYM_STRLEN]=0;

   /* try to reduce strings to character equivalents */
   if (buf[1]!=0 && !StrToChar(buf)) {
      /* still a string, so put it in form (+str) or (-str) */
      if (down) strcpy(buf, "(+");
      else      strcpy(buf, "(-");
      strcat(buf, str);
      strcat(buf, ")");
      return buf;
   }
   if (buf[0]==0) return "";
   if (mod==CONTROL_DOWN) {
      buf[2]=0;
      buf[1]=toupper(buf[0]);
      buf[0]='^';
   }
   if (mod==LOCK_DOWN) buf[0]=toupper(buf[0]);
   return buf;
}


/* returns which modifier is down, shift/caps or control */
int KeyModifiers(char *keys) {

   static int width;
   static XModifierKeymap *mmap;
   int i;

    mmap=XGetModifierMapping(disp);
    width=mmap->max_keypermod;

   for (i=0; i<width; i++) {
      KeyCode code;

      code=mmap->modifiermap[ControlMapIndex*width+i];
      if (code && 0!=BIT(keys, code)) {return CONTROL_DOWN;}

      code=mmap->modifiermap[ShiftMapIndex*width+i];
      if (code && 0!=BIT(keys, code)) {return SHIFT_DOWN;}

      code=mmap->modifiermap[LockMapIndex*width+i];
      if (code && 0!=BIT(keys, code)) {return LOCK_DOWN;}
   }
   return 0;
}

